/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 *      *) validation (getpwent? lh_stats ?).
 *      *) Options at startup.
 **/

#ifndef NONETWORK
#  define NETWORK
#endif

#ifndef NODAEMON
#  define DAEMON
#endif

#include <errno.h> 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>

#include "encoders.h"
#include "command.h"
#include "socket_srv.h"
#include "socket_client.h"
#include "iar.h"
#include "my_stdio.h"
#include "anemometer.h"
#include "coordinate.h"
#include "iar_bus.h"
#include "survey.h"

//int verbose_flag = 1;

int start_survey = 0;

#undef SYSLOG
#define SYSLOG(type, fmt, args...)

#undef PVERB
#ifdef VERBOSE_MODE
#  define PVERB(fmt, args...) if(verbose_flag) fprintf(stderr, fmt, ## args)
#else 
#  define PVERB(fmt, args...)
#endif


#undef PERROR
#define PERROR(s) { PVERB("[ERROR: %s: %s]\n", (s),strerror(errno));\
                    SYSLOG(LOG_ERR,"%s: %s",(s), strerror(errno));\
                  }

#define EXIT(n) exit(n)


void fireman(int s)
{ 
	int wstatus;

	while(wait3(&wstatus, WNOHANG, NULL) > 0);
}


int cartel(int fd, char *prefix)
{
	fd_printf(fd, "%sIAR Antenna position system.\n", prefix);
	fd_printf(fd, "%s\n", prefix);

	return 0;
}


int main(int argc, const char *argv[]) 
{
	int fd_socket, fd_con;
	int fd_pipe, fd_track;
	int pid;
	int ra = RA_CASE;
	int dec = DEC_CASE;

	struct ib_command cm[] = {
		{IB_CM_QUIT, ib_cm_quit},
		{IB_CM_START, ib_cm_start},
		{IB_CM_STOP, ib_cm_stop},
		{IB_CM_MESSAGE, ib_cm_message},
		{IB_CM_WHO, ib_cm_who},
		{IB_CM_MESSAGE_TO, ib_cm_message_to},
		{IB_CM_WHOAMI, ib_cm_whoami},
		{IB_CM_SID, ib_cm_set_id},
		{NULL, NULL}
	};

	if ( (fd_pipe = encoders_open(ra, dec)) == -1){
                perror("main");
                exit(1);
        }

#ifdef DAEMON
	if ( (pid=fork()) != 0 ) {
		exit(0);
	}
#endif

	if ( (pid=fork()) == 0 ) {
		//init_iar_bus_local(cm, fd_pipe, TRACK_POS_SOCK_NAME);
		init_iar_bus_inet(cm, fd_pipe, TRACK_POS_PORT);
	}
	close(fd_pipe);

/* disparar el proceso del survey */
/* FIXME: ... */
	if ( (pid=fork()) == 0 ) {
		//fd_track = call_socket_local(TRACK_POS_SOCK_NAME);
		fd_track = call_socket("127.0.0.1", TRACK_POS_PORT);
		survey_main(fd_track);		
		close(fd_track);
		exit(0);
	}


#ifdef NETWORK
        if ( (fd_socket = socket_init( ANTENNA_TCP_PORT )) < 0) {
                /* perror("socket_init"); */
                exit(1);
        }
#endif
	if (signal(SIGCHLD, fireman) == SIG_ERR) {/* this eliminates zombies */
		PERROR("Signal Failed (SIGCHLD).");
		EXIT(0);
	}
	/*
	if (signal(SIGINT, cbrk_handler) == SIG_ERR) {
		PERROR("Signal Failed (SIGINT).");
		EXIT(0);
	}
	if (signal(SIGTERM, cbrk_handler) == SIG_ERR) {
		PERROR("Signal Failed (SIGTERM).");
		EXIT(0);
	}
	*/

	for(;1;){
#ifdef NETWORK
		/* get a connection */
		if ((fd_con = socket_wait(fd_socket)) < 0) { 
			if (errno == EINTR) 
				/* EINTR might happen on accept(), */
				continue; /* try again */
			perror("accept"); /* bad */
			exit(1);
		}
#endif

#ifdef NOFORK
		pid = 0;
#else
		pid = fork();
#endif
		switch(pid) {            /* try to handle connection */
		case -1 :                /* bad news.  scream and die */
			perror("fork");
			close(fd_socket);
			close(fd_con);
			exit(1);
		case 0 : /* we're the child, do command proses */
			cartel(fd_con,"# ");
			//fd_track = call_socket_local(TRACK_POS_SOCK_NAME);
			fd_track = call_socket("127.0.0.1", TRACK_POS_PORT);
			command_main(fd_con, fd_con, fd_track);
			close(fd_con);
			close(fd_track);
			exit(0);
		default :                 /* we're the parent so look for */
			close(fd_con);    /* another connection */
			continue;
		}
	}

	return 0;
}

