/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 **/

#ifndef _IB_H
#define _IB_H

#include <stdint.h> /* for uintXX_t definition */
#include "cmctl.h"

#define IB_PK_ACTION(a) (a & 0x000f)
#define IB_PK_ADDR(a)  ((a & 0xfff0) >> 4)

#define IB_ACTION_PK(a) (a & 0x000f)
#define IB_ADDR_PK(a)  ((a & 0x0fff) << 4)

#define IB_PACKET_DATA_LEN 80

#define IB_NULL_ADDR      0x0000
#define IB_BUS_ADDR       IB_PK_ADDR(0x8000)
#define IB_BASE_ADDR      IB_PK_ADDR(0x4000)
#define IB_BROADCAST_ADDR IB_PK_ADDR(0xbff0)

/* for status in struct ib_dev */
#define IB_DEV_ON    0x0001
#define IB_DEV_OFF   ( ~ IB_DEV_ON )
#define IB_DEV_READ  0x0004
#define IB_DEV_WRITE 0x0008
#define IB_DEV_RW   ( IB_DEV_READ | IB_DEV_WRITE )
#define IB_DEV_NEWPK 0x0010 /* new packet for dispatching */
#define IB_DEV_ALIVE 0x0020


struct ib_packet_head {
	uint16_t addr; /* To */ 
	uint16_t size;
};


struct ib_packet {
	struct ib_packet_head h;
	void *data;
};


struct ib_dev {
	int fd;     /* socker descriptor */
	int addr;   /* device address */
	int status; /* device on/off, etc ... */
	struct ib_packet pk; /* last packet send from device */
	struct ib_dev *prev;
	struct ib_dev *next;
};


struct ib_command {
	int cm_id;
	int (*func)(struct ib_dev *dev, struct ib_dev *curren);
};


struct ib_data_head {
	uint16_t addr_from;
	uint16_t size;

};

struct ib_data {  /* from de devece side */
	struct ib_data_head h;
	void *data;
};


#define IB_CM_REMOVE     0x1
#define IB_CM_ON         0x2
#define IB_CM_OFF        0x3
#define IB_CM_MESSAGE    0x4

#define IB_CM_MESSAGE_TO 0xf
//#define IB_CM_SET_ADDR   0x6
//#define IB_CM_WHOAMI     0x7
//#define IB_CM_SCAN       0x8
// free: 0x9 ... 0xf


int ib_cm_remove(struct ib_dev *dev, struct ib_dev *current);
int ib_cm_on(struct ib_dev *dev, struct ib_dev *current);
int ib_cm_off(struct ib_dev *dev, struct ib_dev *current);
int ib_cm_message(struct ib_dev *dev, struct ib_dev *current);
//int ib_cm_scan(struct ib_dev *dev, struct ib_dev *current);
int ib_cm_message_to(struct ib_dev *dev, struct ib_dev *current);
//int ib_cm_whoami(struct ib_dev *dev, struct ib_dev *current);
int ib_cm_set_addr(struct ib_dev *dev, struct ib_dev *current);

int init_iar_bus_local(struct ib_command *cm, char *sock_name );
int init_iar_bus_inet(struct ib_command *cm, int port );


int ibus_send( int fd, int addr_to, int action, void *data, int len);

#endif  /* ifndef _IB_H */

