/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 *
 **/

#include <errno.h> 
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "socket_srv.h"
#include "my_stdio.h"
#include "iar_bus.h"


#define BAD_CMD -1
#define BAD_PARAMETER -2
#define BAD_CLIENT -3



#define PRINT_FMT_MESSAGE(to, from, buff) \
                         (fd_printf(to,"%s\n", buff))


struct ib_dev *init_ib_dev()
{
	struct ib_dev *cl;

	cl = (struct ib_dev *) malloc(sizeof(struct ib_dev));
	cl->fd = -1;
	cl->transfer = 0;
	cl->alive = 1;
	cl->id = NULL;
	//cl->cmd_ind = 0;
	cl->mesg = NULL;
	cl->next = cl->prev = NULL;
	
	return cl;
}


int ib_dev_addr(struct ib_dev *cl, int id)
{
	if (id == 0)	
		cl->addr = IB_BASE_ADDR | cl->fd ;
	else 
		cl->addr = id;

	return 0;
}


int free_ib_dev(struct ib_dev *cl)
{
	if(cl->id != NULL) free(cl->id);
	free(cl);

	return 0;
}


struct ib_dev *find_ib_clien (struct ib_dev *cl, char *to_id)
{
	struct ib_dev *clp = cl;
	
	while (clp->next != NULL && strcmp(clp->id, to_id) != 0){
		clp = clp->next;
	}
	if (clp->next == NULL) clp = NULL;
	return clp;
}


int print_client_data(char *buff, char *from, struct ib_dev *cl, int all)
{
	struct ib_dev *clp = cl;
	int l;

	l = strlen(buff)-1;
	if (buff[l] == '\n') buff[l] = 0;
	while(clp->next != NULL){
		if ( (clp->fd) && 
		     (clp->transfer || all) && 
		     (clp->alive) &&
		     (strcmp(clp->id, from) != 0) ) {
			if( !PRINT_FMT_MESSAGE(clp->fd, from, buff) ){
				perror("write client");
				clp->alive = 0;
			}
		}
		clp = clp->next;
	}
	
	return 0;
}


static struct ib_command *find_ib_command(struct ib_command *cm, int cm_id)
{
	int i;

	for (i = 0; cm[i].cm_id; i++)
		if (cm[i].cm_id == cm_id )
			return (&cm[i]);
	
	return ((struct ib_command *)NULL);
}


/* Execute a command. */
static int execute_command(struct ib_command *cm, struct ib_dev *cl, 
			   struct ib_dev *current)
{
	struct ib_command *command;
	
	command = find_ib_command (cm, current->cmd.cm_id);

	if (!command) {
		return BAD_CMD;
	}

	/* Call the function. */
	return ((*(command->func)) (cl, current));
}



int command_from_client(struct ib_command *cm, 
			struct ib_dev *cl, 
			struct ib_dev *clp)
{
	//struct ib_cm_line c;
	
	if ( read(clp->fd, &clp->cmd, 
		  sizeof(struct ib_cm_line)) <= 0) { /* Ops... die */
		printf("read client fail\n");
		close(clp->fd);
		clp->alive = 0;
		return -1;
	}
	
	switch( execute_command(cm, cl, clp) ){
	case 0:
		break;
	default:
	case BAD_CMD:
		fd_printf(clp->fd,"cmd error\n");
		break;
	case BAD_PARAMETER:
		fd_printf(clp->fd,"bad parameter\n");
			break;
	case BAD_CLIENT:
		fd_printf(clp->fd,"bad client name\n");
		break;
	}

	return 0;
}




#if 0
void intHandler (int signo) 
{
      fprintf(stderr, "exit on signal %d\n", signo);
      exit(0);
}
#endif


int init_iar_bus_main(struct ib_command *cm, int fd_main, int fd_socket )
{
	struct ib_dev *cl, *clp, *cld;
	fd_set rfds;
	struct timeval tv;
	int retval;
	char buff[PIPE_BUFFER_LEN];
	int l;
	struct sockaddr isa;
	int addrlen;

	addrlen = sizeof(isa); /* find socket's address for accept() */
	getsockname(fd_socket, (struct sockaddr *)&isa, &addrlen);

#if 0
	if (signal(SIGSEGV, intHandler) == SIG_ERR) {
		perror("Signal Failed (SIGSEGV).");
		exit(0);
	}
#endif
	cl = init_ib_dev();
	for(;1;){
		clp = cl ;
		FD_ZERO(&rfds);
		FD_SET(fd_socket, &rfds);
		if (fd_main >= 0) FD_SET(fd_main, &rfds);
		while(clp->next != NULL) {
			if (clp->fd && clp->alive ){ 
				FD_SET(clp->fd, &rfds);
				clp = clp->next;
			}
			else if ( !clp->alive ) { 
                                /* OK: client dead, cleaning */ 
				cld = clp;
				if (clp->prev != NULL){
					clp = cld->prev;
					clp->next = cld->next;
					cld->next->prev = clp;
					clp = clp->next;
				}else {
					clp = cl = clp->next;
					cl->prev = NULL;
				}
				free_ib_dev(cld);		
			}
		}
		tv.tv_sec = 10;
		tv.tv_usec = 0;
		/* FIXME: 50 ? implement this dynamic */
		retval = select(50, &rfds, NULL, NULL, &tv);
		if (retval) {
			clp = cl ;
			while(clp->next != NULL) {
				if( FD_ISSET(clp->fd, &rfds) && clp->alive ) {
					command_from_client(cm, cl, clp); 
				}
				clp = clp->next;
			}
			if (FD_ISSET(fd_socket, &rfds)) {
				clp = cl ;
				while(clp->next != NULL) clp = clp->next;
				clp->fd = accept(fd_socket, 
						 (struct sockaddr *)&isa, 
						 &addrlen); 
				if( clp->fd ){
					clp->alive = 1;
					printf("connected\n");
					clp->next = init_ib_dev();
					clp->next->prev = clp;
					ib_dev_addr(clp, 0);
				}
			}
			if ((fd_main >= 0) && FD_ISSET(fd_main, &rfds)) {
				l = read(fd_main, buff, PIPE_BUFFER_LEN);
				if (l > 0) {
					buff[l-1] = 0;
					print_client_data(buff, 
							  IB_MAIN_ID, cl, 0);
				}
			}
		}
	}

	return 0;
}


int init_iar_bus_local(struct ib_command *cm, 
			    int fd_main, char *sock_name )
{
	int fd_socket;

	if ( (fd_socket = socket_init_local( sock_name )) < 0) {
		return -1;
	}

	return init_iar_bus_main(cm, fd_main, fd_socket );
}


int init_iar_bus_inet(struct ib_command *cm, int fd_main, int port )
{
	int fd_socket;

	if ( (fd_socket = socket_init( port )) < 0) {
		return -1;
	}

	return init_iar_bus_main(cm, fd_main, fd_socket);
}


/****************************************************************************/

int ib_cm_quit(struct ib_dev *cl, struct ib_dev *current)
{
	close(current->fd);
	current->alive = 0;

	return 0;
}


int ib_cm_start(struct ib_dev *cl, struct ib_dev *current)
{
	current->transfer = 1;

	return 0;
}


int ib_cm_stop(struct ib_dev *cl, struct ib_dev *current)
{
	current->transfer = 0;

	return 0;
}


int ib_cm_message(struct ib_dev *cl, struct ib_dev *current)
{
	print_client_data(current->mesg, current->id, cl, 1);
	current->mesg = NULL;

	return 0;
}


int ib_cm_who(struct ib_dev *cl, struct ib_dev *current)
{

	struct ib_dev *clp = cl;

	while(clp->next != NULL){
		if ( current->fd ) {
			if( !fd_printf(current->fd,"%s: %1d %1d\n", 
				       clp->id, clp->alive, clp->transfer) ){
				perror("write client");
				current->alive = 0;
			}
		}
		clp = clp->next;
	}
	
	return 0;
}


int ib_cm_message_to(struct ib_dev *cl, struct ib_dev *current)
{
	int i;
	char *to_id;
	char *line = current->mesg;
	struct ib_dev *clp;

	i = 0;
	while (line[i] && whitespace (line[i]))	i++;
	to_id = line + i;
	while (line[i] && !whitespace (line[i])) i++;
	if (line[i])
		line[i++] = '\0';

	clp = find_ib_clien(cl, to_id);

	if (!clp) {
		return BAD_CLIENT;
	}

	while (whitespace (line[i])) i++;
	current->mesg = line + i;

	PRINT_FMT_MESSAGE(clp->fd, current->id, current->mesg);
	current->mesg = NULL;

	return 0;
}


int ib_cm_whoami(struct ib_dev *cl, struct ib_dev *current)
{
	fd_printf(current->fd,"%s: %1d %1d\n", 
		  current->id, current->alive, current->transfer);
		
	return 0;
}


#if 0 /*FIXME */
int ib_cm_set_addr(struct ib_dev *cl, struct ib_dev *current)
{

	if( ((*current->mesg > '0') && (*current->mesg < '9')) ||
	    (find_ib_clien(cl, current->mesg)) ){
		return BAD_PARAMETER;
	}
	    
	ib_dev_id(current, current->mesg);
	current->mesg = NULL;
		
	return 0;
}
#endif

/*****************************************************************************
 * Auxiliary functions for messages manager:
 ****************************************************************************/


#if 0 /* FIXME: eliminar */
int skip_prompt(char *buff, char **ptrbuff)
{
	int l, bl;
	
	bl = strlen(buff);
	l = 0;

	*ptrbuff = buff+l;
       
	return bl - l ;
}
#endif

int parse_ib_line(char *line, char **from, char **ptrdata, int *cmd)
{
	int l, bl;
	
	bl = strlen(line);
	l = 0;
#ifndef NO_IB_PROMPT
	while ( *(line+l) != IB_ID_PROPT && l != bl) l++;
	if ( l == bl ) {
		*ptrdata = NULL;
		*from = NULL;
		*cmd = -1;
		return -1;
	}
	*(line+l) = 0;
	l++;
	while(whitespace( *(line+l) )) l++;
	*from = line;
#else
	*from = NULL;
#endif
	*cmd = (int) strtol(line+l, NULL/*ptrdata*/, 16 );
	l += 5;
	*ptrdata = line+l; 
	//fprintf(stderr,"From: [%s] %s\n", *from, *ptrdata);

	return bl - l;
}


#if 0
int ib_cmctl(int fd, char *to, int cmd, ...)
{
	/* FIXME: if arg is NULL not work */
//	va_list ap;
	int ret;
/*	char *arg = NULL;

	va_start(ap, cmd);
	if( ap != NULL )
		arg = va_arg(ap, char *);
*/	
	if ( to != NULL){
/*
		if( arg )
			ret = fd_printf(fd, IB_CM_MESSAGE_TO" %s %04x %s\n", 
					to, cmd, arg);
		else
*/
		ret = fd_printf(fd, IB_CM_MESSAGE_TO" %s %04x\n", 
				to, cmd);
	}
	else{
/*
		if( arg )
			ret = fd_printf(fd, IB_CM_MESSAGE" %04x %s\n", 
					cmd, arg);
		else
*/
		ret = fd_printf(fd, IB_CM_MESSAGE" %04x\n", 
				cmd);
	}
//	va_end(ap);
	
	return ret;
}
#endif


int ib_cmctl(int fd, char *to, int cmd, ...)
{
	int ret;
	if ( to != NULL){
		ret = fd_printf(fd, IB_CM_MESSAGE_TO" %s %04x\n", to, cmd);
	}
	else{
		ret = fd_printf(fd, IB_CM_MESSAGE" %04x\n", cmd);
	}
	
	return ret;
}
