/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 *
 **/

#include <errno.h> 
#include <stdio.h>
#include <stdlib.h>
//#include <stdarg.h>
#include <unistd.h>
//#include <sys/types.h>
//#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "socket_srv.h"
//#include "my_stdio.h"
#include "iar_bus.h"

#define BAD_CMD -1
//#define BAD_PARAMETER -2
//#define BAD_CLIENT -3


int ret_write;


struct ib_dev *init_ib_dev()
{
	struct ib_dev *dev;

	dev = (struct ib_dev *) malloc(sizeof(struct ib_dev));
	dev->pk.data =  (void *) malloc(IB_PACKET_DATA_LEN);

	dev->fd = -1;
	dev->addr = IB_NULL_ADDR;
	dev->status = IB_DEV_ON;
	dev->status &= ~ IB_DEV_ALIVE;
	dev->next = dev->prev = NULL;
	
	return dev;
}


int set_ib_dev_addr(struct ib_dev *dev, int addr)
{
	if (addr == 0)	
		dev->addr = IB_BASE_ADDR | dev->fd ;
	else 
		dev->addr = IB_BASE_ADDR | addr;

	return 0;
}


int free_ib_dev(struct ib_dev *dev)
{
	if( dev->fd >= 0 )
		close(dev->fd);
	free(dev->pk.data);
	free(dev);

	return 0;
}


struct ib_dev *find_ib_dev (struct ib_dev *dev, int addr)
{
	struct ib_dev *devp = dev;
	
	while (devp->next != NULL && (devp->addr != addr) ){
		devp = devp->next;
	}
	if (devp->next == NULL) devp = NULL;
	return devp;
}


/*
int print_bus_data(struct ib_cm_line *cm, struct ib_dev *cl, int all)
{
	struct ib_dev *clp = cl;	
	
	while(clp->next != NULL){
		if ( (clp->fd) && 
		     ((clp->status & IB_DEV_ON) || all) && 
		     ( clp->status & IB_DEV_ALIVE) &&
		     ( clp->addr != cm->addr_f)  ) {
			if( !PRINT_FMT_MESSAGE(clp->fd, cm) ){
				die_ib_dev(clp);
			}
		}
		clp = clp->next;
	}
	
	return 0;
}
*/


int die_ib_dev(struct ib_dev *devp)
{
	/* FIXME: */ printf("read/write device fail\n");
	close(devp->fd);
	devp->fd = -1;
	devp->status &= ~ IB_DEV_ALIVE;

	return 0;
}



int send_data_to_dev(struct ib_data *d, int to_fd)
{
	if( (ret_write = write(to_fd, d, sizeof(struct ib_data_head))) <= 0 ){
		printf("*** write error ***\n");
		return -1;
	}
	//printf("*** ret: %d\n", ret);
	if( d->h.size > 0 ) {
		if( (ret_write = write(to_fd, d->data, d->h.size)) < 0 ){
			printf("*** write error ***\n");
			return -1;
		}
		//printf("*** ret: %d\n", ret);
	}

	return 0;
}


int send_to_dev(struct ib_dev *from, struct ib_dev *to)
{
	struct ib_data d;

	d.h.addr_from = from->addr;
	d.h.size = from->pk.h.size;
	d.data = from->pk.data;

	if( send_data_to_dev( &d, to->fd) < 0 ){
		die_ib_dev( to );
		return -1;
	}
	/*
	if( write(to->fd, &d, sizeof(struct ib_data_head)) <= 0 ){
		die_ib_dev( to );
		return -1;
	}
	if( d.h.size > 0 ) {
		if( write(to->fd, d.data, d.h.size) < 0 ){
			die_ib_dev( to );
			return -1;
		}
	}
	*/
	return 0;
}


static struct ib_command *find_ib_command(struct ib_command *cm, int cm_id)
{
	int i;

	for (i = 0; cm[i].cm_id; i++) {
		if (cm[i].cm_id == cm_id ) {
			return (&cm[i]);
		}
	}
	
	return ((struct ib_command *)NULL);
}


/* Execute a command. */
static int execute_command(struct ib_command *cm, struct ib_dev *dev,
			   struct ib_dev *devp)
{
	struct ib_command *command;
	
	command = find_ib_command (cm, IB_PK_ACTION(devp->pk.h.addr));

	if (!command) {
		/* FIXME: */ printf("*** BAD CMD *** [0x%x 0x%x]\n", IB_PK_ACTION(devp->pk.h.addr), devp->pk.h.addr );
		return BAD_CMD;
	}
	/*FIXME: */ //printf("cmd find.\n");
	/* Call the function. */
	return ((*(command->func)) (dev, devp));
}


int read_cm_from_dev(struct ib_dev *devp)
{
	//int *v;

	memset(&devp->pk, 0, sizeof(struct ib_packet_head));
	if ( read(devp->fd, &devp->pk.h, sizeof(struct ib_packet_head)) <= 0) {
		/* Ops... die */
		die_ib_dev( devp );
		return -1;
	}
	/* FIXME: */
	//printf("addrto: 0x%x, 0x%x, %d", 
	//      devp->pk.h.addr, IB_PK_ACTION(devp->pk.h.addr), 
	//      devp->pk.h.size);
	
	if (  devp->pk.h.size > 0 && devp->pk.h.size < IB_PACKET_DATA_LEN ){
		memset(devp->pk.data, 0, devp->pk.h.size);
		if ( read(devp->fd, devp->pk.data, devp->pk.h.size) <= 0 ) {
			die_ib_dev( devp );
			return -1;
		}
		//v = devp->pk.data;
		//printf(", data: [%d]", *v );
	}
	
	//printf("\n");
	devp->status |= IB_DEV_NEWPK;

	return 0;
}


int process_cm_queue(struct ib_command *cm, struct ib_dev *dev)
{
	struct ib_dev *devp = dev;

	while(devp != NULL) {
		if ( devp->status & IB_DEV_NEWPK ){
			execute_command(cm, dev, devp);
			devp->status &= ~IB_DEV_NEWPK;
		}
		devp = devp->next;
	}

	return 0;
}


void intHandler (int signo) 
{
	fprintf(stderr, "ret write %d\n", ret_write);
      //exit(0);
	
	return;
}

int iar_bus_core(struct ib_command *cm, int fd_socket )
{
	struct ib_dev *dev, *devp, *devd;
	fd_set rfds;
	struct timeval tv;
	int retval;
	struct sockaddr isa;
	int addrlen;
	int queue;
	int fd_max;

	addrlen = sizeof(isa); /* find socket's address for accept() */
	getsockname(fd_socket, (struct sockaddr *)&isa, &addrlen);

	if (signal(SIGPIPE, intHandler) == SIG_ERR) {
		perror("Signal Failed (SIGPIPE).");
		exit(0);
	}

	dev = init_ib_dev();
	/* bus dev */
	dev->fd = fd_socket;
	dev->addr = IB_BUS_ADDR;
	dev->status = IB_DEV_ON | IB_DEV_ALIVE | IB_DEV_READ;
	/* ... */

	for(;1;){
		devp = dev;
		FD_ZERO(&rfds);
		fd_max = fd_socket + 1;
		while(devp != NULL) {
			if (devp->fd && (devp->status & IB_DEV_ALIVE) ){ 
				FD_SET(devp->fd, &rfds);
				if( fd_max < devp->fd + 1)
					fd_max = devp->fd + 1;
			}
			else if ( !(devp->status & IB_DEV_ALIVE) && 
				  (devp != dev) ) { 
                                /* OK: device dead, cleaning */ 
				devd = devp;
				devp = devd->prev;
				devp->next = devd->next;
				if( devd->next )
					devd->next->prev = devp;
				free_ib_dev(devd);	
			}
			devp = devp->next;
		}
		tv.tv_sec = 5;
		tv.tv_usec = 0;
		retval = select(fd_max, &rfds, NULL, NULL, &tv);
		if (retval) {
			devp = dev->next ; /* skip device bus */
			queue = 0;
			while(devp != NULL) {
				if(devp->status & IB_DEV_ALIVE)
					if( FD_ISSET(devp->fd, &rfds) ) {
						read_cm_from_dev(devp);
						queue = 1;
						//process_cm_queue(cm, dev->next);
					}
				devp = devp->next;
			}
			if( queue )
				process_cm_queue(cm, dev->next);
			if (FD_ISSET(dev->fd, &rfds)) {
				/* New device is connected */
				/* FIXME: */ printf("new call\n");
				devp = dev ;
				while(devp->next != NULL) devp = devp->next;
				devp->next = init_ib_dev();
				devp->next->prev = devp;
				devp = devp->next;
				devp->fd = accept(dev->fd, 
						  (struct sockaddr *)&isa, 
						  &addrlen); 
				if( devp->fd ){
					devp->status |= IB_DEV_ALIVE;
					/* FIXME: */ printf("connected\n");
					set_ib_dev_addr(devp, 0);
				}
			}
		}
	}

	return 0;
}


int init_iar_bus_local(struct ib_command *cm, char *sock_name )
{
	int fd_socket;

	if ( (fd_socket = socket_init_local( sock_name )) < 0) {
		return -1;
	}

	return iar_bus_core(cm, fd_socket );
}


int init_iar_bus_inet(struct ib_command *cm, int port )
{
	int fd_socket;

	if ( (fd_socket = socket_init( port )) < 0) {
		return -1;
	}

	return iar_bus_core(cm, fd_socket);
}


/****************************************************************************/

int ib_cm_remove(struct ib_dev *dev, struct ib_dev *current)
{
	current->status &= ~ IB_DEV_ON;
	current->status &= ~ IB_DEV_ALIVE;

	return close( current->fd );
}


int ib_cm_on(struct ib_dev *dev, struct ib_dev *current)
{
	current->status |= IB_DEV_ON;

	return 0;
}


int ib_cm_off(struct ib_dev *dev, struct ib_dev *current)
{
	current->status &= ~ IB_DEV_ON;

	return 0;
}


int ib_cm_message(struct ib_dev *dev, struct ib_dev *current)
{
	struct ib_dev *devp = dev;

	while (devp != NULL){
		if( (devp->status & IB_DEV_ALIVE) && 
		    (devp != current) ){
			//printf("sending cmd to: %d\n", devp->fd);
			send_to_dev(current, devp);
		}
			devp = devp->next;
	}
	
	return 0;
}


int ib_cm_message_to(struct ib_dev *dev, struct ib_dev *current)
{
	send_to_dev(current, find_ib_dev(dev, IB_PK_ADDR(current->pk.h.addr)));
	
	return 0;
}


int ib_cm_set_addr(struct ib_dev *dev, struct ib_dev *current)
{
	int new_addr = (int) current->pk.data;

	if( find_ib_dev(dev, new_addr) != NULL ) {
		return -1;
	}
	set_ib_dev_addr(current, new_addr);
	
	return 0;
}

